// Code generated by "go run len_generate.go"; DO NOT EDIT.

package dns

func (rr *NULL) Len() int       { return rr.Hdr.Len() + rr.NULL.Len() }
func (rr *CNAME) Len() int      { return rr.Hdr.Len() + rr.CNAME.Len() }
func (rr *HINFO) Len() int      { return rr.Hdr.Len() + rr.HINFO.Len() }
func (rr *MB) Len() int         { return rr.Hdr.Len() + rr.MB.Len() }
func (rr *MG) Len() int         { return rr.Hdr.Len() + rr.MG.Len() }
func (rr *MINFO) Len() int      { return rr.Hdr.Len() + rr.MINFO.Len() }
func (rr *MR) Len() int         { return rr.Hdr.Len() + rr.MR.Len() }
func (rr *MF) Len() int         { return rr.Hdr.Len() + rr.MF.Len() }
func (rr *MD) Len() int         { return rr.Hdr.Len() + rr.MD.Len() }
func (rr *MX) Len() int         { return rr.Hdr.Len() + rr.MX.Len() }
func (rr *AFSDB) Len() int      { return rr.Hdr.Len() + rr.AFSDB.Len() }
func (rr *X25) Len() int        { return rr.Hdr.Len() + rr.X25.Len() }
func (rr *ISDN) Len() int       { return rr.Hdr.Len() + rr.ISDN.Len() }
func (rr *RT) Len() int         { return rr.Hdr.Len() + rr.RT.Len() }
func (rr *NS) Len() int         { return rr.Hdr.Len() + rr.NS.Len() }
func (rr *PTR) Len() int        { return rr.Hdr.Len() + rr.PTR.Len() }
func (rr *RP) Len() int         { return rr.Hdr.Len() + rr.RP.Len() }
func (rr *SOA) Len() int        { return rr.Hdr.Len() + rr.SOA.Len() }
func (rr *TXT) Len() int        { return rr.Hdr.Len() + rr.TXT.Len() }
func (rr *IPN) Len() int        { return rr.Hdr.Len() + rr.IPN.Len() }
func (rr *SRV) Len() int        { return rr.Hdr.Len() + rr.SRV.Len() }
func (rr *NAPTR) Len() int      { return rr.Hdr.Len() + rr.NAPTR.Len() }
func (rr *CERT) Len() int       { return rr.Hdr.Len() + rr.CERT.Len() }
func (rr *DNAME) Len() int      { return rr.Hdr.Len() + rr.DNAME.Len() }
func (rr *A) Len() int          { return rr.Hdr.Len() + rr.A.Len() }
func (rr *AAAA) Len() int       { return rr.Hdr.Len() + rr.AAAA.Len() }
func (rr *PX) Len() int         { return rr.Hdr.Len() + rr.PX.Len() }
func (rr *GPOS) Len() int       { return rr.Hdr.Len() + rr.GPOS.Len() }
func (rr *LOC) Len() int        { return rr.Hdr.Len() + rr.LOC.Len() }
func (rr *RRSIG) Len() int      { return rr.Hdr.Len() + rr.RRSIG.Len() }
func (rr *DS) Len() int         { return rr.Hdr.Len() + rr.DS.Len() }
func (rr *KX) Len() int         { return rr.Hdr.Len() + rr.KX.Len() }
func (rr *TA) Len() int         { return rr.Hdr.Len() + rr.TA.Len() }
func (rr *TALINK) Len() int     { return rr.Hdr.Len() + rr.TALINK.Len() }
func (rr *SSHFP) Len() int      { return rr.Hdr.Len() + rr.SSHFP.Len() }
func (rr *DNSKEY) Len() int     { return rr.Hdr.Len() + rr.DNSKEY.Len() }
func (rr *RKEY) Len() int       { return rr.Hdr.Len() + rr.RKEY.Len() }
func (rr *NSAPPTR) Len() int    { return rr.Hdr.Len() + rr.NSAPPTR.Len() }
func (rr *NSEC3PARAM) Len() int { return rr.Hdr.Len() + rr.NSEC3PARAM.Len() }
func (rr *TKEY) Len() int       { return rr.Hdr.Len() + rr.TKEY.Len() }
func (rr *RFC3597) Len() int    { return rr.Hdr.Len() + rr.RFC3597.Len() }
func (rr *URI) Len() int        { return rr.Hdr.Len() + rr.URI.Len() }
func (rr *DHCID) Len() int      { return rr.Hdr.Len() + rr.DHCID.Len() }
func (rr *TLSA) Len() int       { return rr.Hdr.Len() + rr.TLSA.Len() }
func (rr *SMIMEA) Len() int     { return rr.Hdr.Len() + rr.SMIMEA.Len() }
func (rr *HIP) Len() int        { return rr.Hdr.Len() + rr.HIP.Len() }
func (rr *NINFO) Len() int      { return rr.Hdr.Len() + rr.NINFO.Len() }
func (rr *NID) Len() int        { return rr.Hdr.Len() + rr.NID.Len() }
func (rr *L32) Len() int        { return rr.Hdr.Len() + rr.L32.Len() }
func (rr *L64) Len() int        { return rr.Hdr.Len() + rr.L64.Len() }
func (rr *LP) Len() int         { return rr.Hdr.Len() + rr.LP.Len() }
func (rr *EUI48) Len() int      { return rr.Hdr.Len() + rr.EUI48.Len() }
func (rr *EUI64) Len() int      { return rr.Hdr.Len() + rr.EUI64.Len() }
func (rr *CAA) Len() int        { return rr.Hdr.Len() + rr.CAA.Len() }
func (rr *UID) Len() int        { return rr.Hdr.Len() + rr.UID.Len() }
func (rr *GID) Len() int        { return rr.Hdr.Len() + rr.GID.Len() }
func (rr *UINFO) Len() int      { return rr.Hdr.Len() + rr.UINFO.Len() }
func (rr *EID) Len() int        { return rr.Hdr.Len() + rr.EID.Len() }
func (rr *NIMLOC) Len() int     { return rr.Hdr.Len() + rr.NIMLOC.Len() }
func (rr *OPENPGPKEY) Len() int { return rr.Hdr.Len() + rr.OPENPGPKEY.Len() }
func (rr *ZONEMD) Len() int     { return rr.Hdr.Len() + rr.ZONEMD.Len() }
func (rr *SVCB) Len() int       { return rr.Hdr.Len() + rr.SVCB.Len() }
func (rr *DELEG) Len() int      { return rr.Hdr.Len() + rr.DELEG.Len() }
func (rr *DSYNC) Len() int      { return rr.Hdr.Len() + rr.DSYNC.Len() }
func (rr *TSIG) Len() int       { return rr.Hdr.Len() + rr.TSIG.Len() }
