// Code generated by "go run dsolen_generate.go"; DO NOT EDIT.

package dns

func (rr *KEEPALIVE) Len() int {
	l := 0
	l += 4 // Timeout
	l += 4 // Interval
	return l
}

func (rr *RETRYDELAY) Len() int {
	l := 0
	l += 4 // Delay
	return l
}

func (rr *DPADDING) Len() int {
	l := 0
	l += len(rr.Padding) / 2
	return l
}
